/* $Id: hstream.c,v 1.20 2002/10/07 07:34:49 aa5779 Exp $ */
/*
	Copyright (C) 2001, 2002 Artem V. Andreev (artem@AA5779.spb.edu)

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <ctype.h>
#include <errno.h>
#include <limits.h>
#include <string.h>
#include "util.h"
#include "lists.h"
#include "interp.h"
#define HSTREAM_IMP
#include "hstream.h"
#undef getc
#undef putc
#undef puts

stream_class html_stream_class = {"html", hs_open, hs_close, hs_read, hs_writestr, 
								   hs_writetag, hs_ioctl, bs_getpos,
								   bs_setpos, bs_rewind
								 };

stream *hs_open(interpreter *i, char *name, char *fname, int writeable)
{
	stream *ns = create_basic(name, fname, sizeof(html_stream), writeable);
	html_stream *hs = ns->data;
	hs->is_7bit = 0;
	hs->allow_tags = 1;
	hs->allow_ents = 1;
	hs->pack_spaces = 1;
	hs->upper_case = 1;
	hs->pending = 0;
	ns->class = &html_stream_class;
	return ns;
}

void hs_close(interpreter *i, stream *s)
{
	html_stream *hs = s->data;
	release_html_chunk(&hs->current);
	bs_close(i, s);
}

static void next_html_chunk(html_stream *hs);
html_chunk *hs_read(interpreter *i, stream *s)
{
	html_stream *hs = s->data;
	if(!hs->pending)
	{
		if(!hs->basic.cooked_mode)
			return bs_read(i, s);
		next_html_chunk(hs);
	}
	hs->pending = 0;
	return HTML_CHUNK_TYPE(&hs->current) != hct_none ? &hs->current : NULL;
}

void hs_writestr(interpreter *i, stream *s, char *str)
{
	html_stream *hs = s->data;
	if(s->direction != stream_out)
		return;
	if(!hs->basic.cooked_mode)
	{
		bs_writestr(i, s, str);
		return;
	}
	for(; *str; str++)
	{
		switch(*str)
		{
			case '<':
				hs->basic.fc->puts(hs->basic.fh, "&lt;");
				break;
			case '>':
				hs->basic.fc->puts(hs->basic.fh, "&gt;");
				break;
			case '"':
				hs->basic.fc->puts(hs->basic.fh, "&quot;");
				break;
			case '&':
				if(hs->allow_ents)
				{
					hs->basic.fc->puts(hs->basic.fh, "&amp;");
					break;
				}
			default:
				if(hs->is_7bit && (*str & 0x80))
				{
					char buf[16];
					sprintf(buf, "&%d;", *(unsigned char *)str);
					hs->basic.fc->puts(hs->basic.fh, buf);
				}
				else hs->basic.fc->putc(hs->basic.fh, *str);
		}
	}
}

static int is_digits_only(char *str)
{
	while(*str)
	{
		if(!isdigit(*str++))
			return 0;
	}
	return 1;
}

void hs_writetag(interpreter *i, stream *s, html_chunk *ch)
{
	html_stream *hs = s->data;
	if(ch->type == hct_char)
	{
		char buf[2] = {0, 0};
		buf[0] = ch->x.ch;
		s->class->writestr(i, s, buf);
	}
	else if(*get_tag_name(ch) == '&' && !get_tag_name(ch)[1])
	{
		char *v;
		hs->basic.fc->putc(hs->basic.fh, '&');
		v = get_html_param(ch, "value");
		if(is_digits_only(v))
			hs->basic.fc->putc(hs->basic.fh, '#');
		hs->basic.fc->puts(hs->basic.fh, v);
		hs->basic.fc->putc(hs->basic.fh, ';');
	}
	else if(*get_tag_name(ch) == '!' && !get_tag_name(ch)[1])
	{
		hs->basic.fc->putc(hs->basic.fh, '<');
		hs->basic.fc->putc(hs->basic.fh, *get_tag_name(ch));
		hs->basic.fc->puts(hs->basic.fh, get_html_param(ch, "rest"));
		hs->basic.fc->putc(hs->basic.fh, '>');
	}
	else if(*get_tag_name(ch) == '!' && get_tag_name(ch)[1] == '[' && !get_tag_name(ch)[2])
	{
		hs->basic.fc->puts(hs->basic.fh, "<![");
		hs->basic.fc->puts(hs->basic.fh, get_html_param(ch, "SECTION"));
		hs->basic.fc->putc(hs->basic.fh, '[');
	}
	else if(*get_tag_name(ch) == '?' && !get_tag_name(ch)[1])
	{
		char *target = get_html_param(ch, "TARGET");
		if(target)
		{
			hs->basic.fc->puts(hs->basic.fh, "<?");
			hs->basic.fc->puts(hs->basic.fh, target);
			hs->basic.fc->putc(hs->basic.fh, ' ');
			hs->basic.fc->puts(hs->basic.fh, get_html_param(ch, "rest"));	
			hs->basic.fc->puts(hs->basic.fh, "?>");
		}
		else
		{
			hs->basic.fc->puts(hs->basic.fh, "<?");
			hs->basic.fc->puts(hs->basic.fh, get_html_param(ch, "rest"));
			hs->basic.fc->putc(hs->basic.fh, '>');
		}
	}
	else if(!strcmp(get_tag_name(ch), "!--"))
	{
		hs->basic.fc->puts(hs->basic.fh, "<!-- ");
		hs->basic.fc->puts(hs->basic.fh, get_html_param(ch, "comment"));
		hs->basic.fc->puts(hs->basic.fh, "-->");
	}
	else
	{
		html_param *parm;
		char *concur;
		hs->basic.fc->putc(hs->basic.fh, '<');
		if((concur = get_html_param(ch, "!CONCUR"))) // SGML CONCUR support
		{
			hs->basic.fc->putc(hs->basic.fh, '(');
			hs->basic.fc->puts(hs->basic.fh, concur);
			hs->basic.fc->putc(hs->basic.fh, ')');
		}
		hs->basic.fc->puts(hs->basic.fh, get_tag_name(ch));
		for(parm = ch->params; parm; parm = list_next(parm))
		{
			if(!ds_comparestr(parm->name, "!CONCUR", ds_p_casefold, NULL) ||
			   !ds_comparestr(parm->name, "!malformed!", ds_p_casefold, NULL)
			  )
				continue;
			hs->basic.fc->putc(hs->basic.fh, ' ');
			hs->basic.fc->puts(hs->basic.fh, DS_BODY(parm->name));
			if(parm->val && ds_compare(parm->name, parm->val, NULL, NULL))
			{
				hs->basic.fc->puts(hs->basic.fh, "=\"");
				hs->basic.fc->puts(hs->basic.fh, DS_BODY(parm->val));
				hs->basic.fc->putc(hs->basic.fh, '\"');
			}
		}
		concur = get_html_param(ch, "!malformed!"); // concur reused just as a temp var
		if(concur)
			hs->basic.fc->puts(hs->basic.fh, concur);
		hs->basic.fc->putc(hs->basic.fh, '>');
	}
}

char *hs_ioctl(interpreter *i, stream *s, char *ctlstr)
{
	html_stream *hs = s->data;

	if(!strcasecmp(ctlstr, "unread"))
	{
		hs->pending = 1;
		return "";
	}
	else if(!strcmp(ctlstr, "7bit"))
	{
		hs->is_7bit = 1;
		return "";
	}
	else if(!strcmp(ctlstr, "8bit"))
	{
		hs->is_7bit = 0;
		return "";
	}
	else if(!strcmp(ctlstr, "?bit"))
	{
		return hs->is_7bit ? "7bit" : "8bit";
	}
	else if(!strcmp(ctlstr, "usetags"))
	{
		hs->allow_tags = 1;
		return "";
	}
	else if(!strcmp(ctlstr, "nousetags"))
	{
		hs->allow_tags = 0;
		return "";
	}
	else if(!strcmp(ctlstr, "?usetags"))
	{
		return hs->allow_tags ? "usetags" : "nousetags";
	}
	else if(!strcmp(ctlstr, "useents"))
	{
		hs->allow_ents = 1;
		return "";
	}
	else if(!strcmp(ctlstr, "nouseents"))
	{
		hs->allow_ents = 0;
		return "";
	}
	else if(!strcmp(ctlstr, "?useents"))
	{
		return hs->allow_ents ? "useents" : "nouseents";
	}
	else if(!strcmp(ctlstr, "keepspaces"))
	{
		hs->pack_spaces = 0;
		return "";
	}
	else if(!strcmp(ctlstr, "nokeepspaces"))
	{
		hs->pack_spaces = 1;
		return "";
	}
	else if(!strcmp(ctlstr, "?keepspaces"))
	{
		return !hs->pack_spaces ? "keepspaces" : "nokeepspaces";
	}
	else if(!strcmp(ctlstr, "uppercase"))
	{
		hs->upper_case = 1;
		return "";
	}
	else if(!strcmp(ctlstr, "nouppercase"))
	{
		hs->upper_case = 0;
		return "";
	}
	else if(!strcmp(ctlstr, "?uppercase"))
	{
		return hs->upper_case ? "uppercase" : "nouppercase";
	}
	return bs_ioctl(i, s, ctlstr);
}

static void parse_compound_char(html_stream *hs);
static void parse_html_tag(html_stream *hs);
static void next_html_chunk(html_stream *hs)
{
	int ch, skip = 0;
	
	reset_html_chunk(&hs->current);
	ch = hs->basic.fc->getc(hs->basic.fh);
   	switch(ch)
    {
		case FLOW_EOF:
			break;
		case FLOW_ERROR:
			recover_error("flow read error: %s", strerror(errno));
			break;
		case '&':
			if(!hs->allow_ents)
				skip = 1;
			else
			{
				parse_compound_char(hs);
				break;
			}
		case '<':
			if(hs->allow_tags && !skip)
			{
				parse_html_tag(hs);
				break;
			}
		default:
			if(!isspace(ch) || !hs->pack_spaces)
				set_html_char(&hs->current, ch);
			else
			{
				set_html_char(&hs->current, ' ');
				while(isspace(ch))
					ch = hs->basic.fc->getc(hs->basic.fh);
				if(ch != FLOW_EOF)
					hs->basic.fc->ungetc(hs->basic.fh, ch);
			}
	}	
}

typedef struct name_code
{
    char *name;
    int code;
} name_code;

static int lookup_char_name(char *name)
{
    static name_code name_table[] = { {"amp", '&'}, {"lt", '<'}, {"gt", '>'}, {"quot", '"'}, 
									  {"apos", '\''}, {NULL, 0}};
    name_code *niter;
    for(niter = name_table; niter->name; niter++)
    {
		if(!strcmp(niter->name, name))
	    	break;
    }
    return niter->code;
}

static void parse_compound_char(html_stream *hs)
{
	dstring *name = ds_create(NULL);
	int ch;

	ch = hs->basic.fc->getc(hs->basic.fh);
    if(ch == '#')
   	{
		ch = hs->basic.fc->getc(hs->basic.fh);
		if(ch == 'x' || ch == 'X') 
		{
			int val = 0;
			for(;;)
			{
				ch = hs->basic.fc->getc(hs->basic.fh);
				if(!isxdigit(ch))
					break;
				val = (val << 4) | (isdigit(ch) ? ch - '0' : toupper(ch) - 'A' + 10);
			}
			name = ds_fromint(val, 10, 0);;
		}
		else
		{
			while(isdigit(ch))
			{
				ds_appendch(name, ch);
				ch = hs->basic.fc->getc(hs->basic.fh);
			}
		}
	}
    else
    {
		while(isalnum(ch))
		{
	    	ds_appendch(name, ch);
			ch = hs->basic.fc->getc(hs->basic.fh);
		}
		if(ds_isempty(name))
			set_html_char(&hs->current, '&');
		else
		{
			int ch1 = lookup_char_name(DS_BODY(name));
			if(ch1)
				set_html_char(&hs->current, ch1);
		}
	}
	if(HTML_CHUNK_TYPE(&hs->current) == hct_none)
	{
		set_tag_name(&hs->current, ds_create("&"));
		set_html_param(&hs->current, ds_create("value"), name);
	}
	if(ch != ';')
		hs->basic.fc->ungetc(hs->basic.fh, ch);
}

static dstring *gather_until_xmlend(html_stream *hs);
static int flow_skip_spaces(html_stream *hs, int ch);
static dstring *gather_until_angle(html_stream *hs);
static void parse_html_tag(html_stream *hs)
{
	dstring *buffer = ds_create(NULL), *temp = NULL, *temp1;
	dstring *concur = NULL;
	int ch;
	
	ch = hs->basic.fc->getc(hs->basic.fh);
	if(ch == '?')
	{
		set_tag_name(&hs->current, ds_create("?"));
		while(isnamesym(ch = hs->basic.fc->getc(hs->basic.fh)))
			 ds_appendch(buffer, ch);
		set_html_param(&hs->current, ds_create("TARGET"), buffer);
		
		set_html_param(&hs->current, ds_create("rest"), gather_until_xmlend(hs));
		return;
	}
	if(ch == '!')
	{
		char buf[3] = {0, 0, 0};
		*buf = hs->basic.fc->getc(hs->basic.fh);
		if(*buf == '[')
		{
			ch = flow_skip_spaces(hs, hs->basic.fc->getc(hs->basic.fh));
			while(!isspace(ch) && ch != '[' && ch != ']')
			{
				ds_appendch(buffer, ch);
				ch = hs->basic.fc->getc(hs->basic.fh);
			}
			ch = flow_skip_spaces(hs, ch);
			if(ch != '[')
				hs->basic.fc->ungetc(hs->basic.fh, ch);
			else
			{
				set_tag_name(&hs->current, ds_create("!["));
				set_html_param(&hs->current, ds_create("SECTION"), buffer);
				return;
			}
		}
		buf[1] = hs->basic.fc->getc(hs->basic.fh);
		if(*buf != '-' || buf[1] != '-')
		{
			ds_appendstr(buffer, buf);
			set_tag_name(&hs->current, ds_create("!"));
			ch = flow_skip_spaces(hs, hs->basic.fc->getc(hs->basic.fh));
			hs->basic.fc->ungetc(hs->basic.fh, ch);
			set_html_param(&hs->current, ds_create("rest"), ds_concat(buffer, gather_until_angle(hs)));
			return;
		}
		else
		{
			ch = hs->basic.fc->getc(hs->basic.fh);
			for(;;)
			{
				while(ch != '-')
				{
					ds_appendch(buffer, ch);
					ch = hs->basic.fc->getc(hs->basic.fh);
				}
				ch = hs->basic.fc->getc(hs->basic.fh);
				if(ch != '-')
					continue;
				ch = hs->basic.fc->getc(hs->basic.fh);
				if(ch == '>')
					break;
			}
			set_tag_name(&hs->current, ds_create("!--"));
			set_html_param(&hs->current, ds_create("comment"), buffer); 
			return;
		}
	}
	if(ch == '(') // SGML CONCUR support
	{
		concur = ds_create(NULL);
		while((ch = hs->basic.fc->getc(hs->basic.fh)) != ')')
			ds_appendch(concur, ch);
		ch = hs->basic.fc->getc(hs->basic.fh);
	}
    if(ch == '/')
	{
	 	ds_appendch(buffer, ch);
		ch = hs->basic.fc->getc(hs->basic.fh);	
	}
    while(isnamesym(ch))
	{
		ds_appendch(buffer, hs->upper_case ? toupper(ch) : ch);
		ch = hs->basic.fc->getc(hs->basic.fh);
	}
	set_tag_name(&hs->current, buffer);
	if(concur)
		set_html_param(&hs->current, ds_create("!CONCUR"), concur);
    for(;;)
    {
		ch = flow_skip_spaces(hs, ch);
		if(ch == '/')
		{
			set_html_param(&hs->current, ds_create("/"), ds_create("/"));
			if(hs->basic.fc->getc(hs->basic.fh) != '>')
				set_html_param(&hs->current, ds_create("!malformed!"), gather_until_angle(hs));
			break;
		}
		if(ch == '>')
			break;
		if(!isnamesym(ch))
		{
			set_html_param(&hs->current, ds_create("!malformed!"),
							ds_concat(ds_createch(ch), 
									  gather_until_angle(hs)
									 )
						  );
			break;
		}
		temp = ds_create(NULL);
        while(isnamesym(ch))
		{
    	    ds_appendch(temp, hs->upper_case ? toupper(ch) : ch);
			ch = hs->basic.fc->getc(hs->basic.fh);
		}
		ch = flow_skip_spaces(hs, ch);
		temp1 = ds_create(NULL);
		if(ch == '=')
		{
			ch = flow_skip_spaces(hs, hs->basic.fc->getc(hs->basic.fh));
	    	if(ch == '"' || ch == '\'')
	    	{
				int delim = ch;
				for(;;)
				{
					ch = hs->basic.fc->getc(hs->basic.fh);
					if(ch == delim)
						break;
					if(ch == FLOW_EOF || ch == FLOW_ERROR)
						recover_error("missing quote in input stream");
		    		ds_appendch(temp1, ch);
				}
				ch = hs->basic.fc->getc(hs->basic.fh);
	    	}
	    	else if(isnamesym(ch))
			{
				for(;;)
				{
					ds_appendch(temp1, ch);
					ch = hs->basic.fc->getc(hs->basic.fh);
					if(!isnamesym(ch))
						break;
				}
			}
			else
		    {
				set_html_param(&hs->current, ds_create("!malformed!"),
								ds_concat(ds_createch(ch), 
										  gather_until_angle(hs)
										 )
							  );
				break;
	    	}
		}
		else ds_assign(temp1, temp);
		set_html_param(&hs->current, temp, temp1);
    }
	return;
}

static int flow_skip_spaces(html_stream *hs, int ch)
{
	while(isspace(ch))
		ch = hs->basic.fc->getc(hs->basic.fh);
	return ch;
}

static void skip_until_angle(html_stream *hs)
{
	int ch;
	int cnt = 1;
	for(;;)
	{
		ch = hs->basic.fc->getc(hs->basic.fh);
		if(ch == '>')
		{
			if(!--cnt)
				break;
		}
		else if(ch == '<')
			cnt++;
	}
}

static dstring *gather_until_angle(html_stream *hs)
{
	int ch;
	int cnt = 1;
	dstring *ds = ds_create(NULL);
	for(;;)
	{
		ch = hs->basic.fc->getc(hs->basic.fh);
		if(ch == '>')
		{
			if(!--cnt)
				break;
		}
		else if(ch == '<')
			cnt++;
		ds_appendch(ds, ch);
	}
	return ds;
}

static dstring *gather_until_xmlend(html_stream *hs)
{
	int ch;
	dstring *ds = ds_create(NULL);
	for(;;)
	{
		ch = hs->basic.fc->getc(hs->basic.fh);
		if(ch == '?')
		{
			ch = hs->basic.fc->getc(hs->basic.fh);
			if(ch == '>')
				break;
			ds_appendch(ds, '?');
		}
		ds_appendch(ds, ch);
	}
	return ds;
}

